import re

from typing import Iterable

from .chapter import Reference, LineLayout, AssetLayout, ParagraphLayout
from .mark import transform2mark, Mark


_START_PREFIX_PATTERN = re.compile(r"^\*{1,6}\s+")

class References:
    def __init__(self, page_index: int, layouts: Iterable[AssetLayout | ParagraphLayout]) -> None:
        self._page_index: int = page_index
        self._references: list[Reference] = list(self._extract_references(page_index, layouts))
        self._mark2reference: dict[str | Mark, Reference] = {}
        for reference in self._references:
            mark = reference.mark
            if mark not in self._mark2reference:
                self._mark2reference[mark] = reference

    @property
    def page_index(self) -> int:
        return self._page_index

    def get(self, mark: str | Mark) -> Reference | None:
        return self._mark2reference.get(mark, None)

    def _extract_references(self, page_index: int, layouts: Iterable[AssetLayout | ParagraphLayout]):
        order: int = 1
        reference: Reference | None = None
        for item in self._iter_and_inject_marks(layouts):
            if isinstance(item, Mark | str):
                if reference:
                    yield reference
                reference = Reference(
                    page_index=page_index,
                    order=order,
                    mark=item,
                    layouts=[],
                )
                order += 1
            elif reference:
                reference.layouts.append(item)
            else:
                # TODO: 多余的内容可能是上一页的跨页页脚注释 / 引用，也可能是必须忽略的多余内容。
                #       此处没有能力进行判断，以后看看有什么好办法。
                pass
        if reference:
            yield reference

    def _iter_and_inject_marks(self, layouts: Iterable[AssetLayout | ParagraphLayout]):
        for layout in layouts:
            if isinstance(layout, AssetLayout):
                yield layout
            elif isinstance(layout, ParagraphLayout):
                for mark, sub_layout in self._split_paragraph_by_marks(layout):
                    if mark is not None:
                        yield mark
                    yield sub_layout

    def _split_paragraph_by_marks(self, to_split_layout: ParagraphLayout):
        mark_layout: tuple[Mark | str | None, ParagraphLayout] = (
            None,
            ParagraphLayout(ref=to_split_layout.ref, lines=[]),
        )
        for line in to_split_layout.lines:
            mark, content = self._extract_mark(line.content)
            if mark is None:
                mark_layout[1].lines.append(line)
            else:
                if mark_layout[1].lines:
                    yield mark_layout
                mark_layout = (mark, ParagraphLayout(
                    ref=to_split_layout.ref,
                    lines=[LineLayout(
                        page_index=line.page_index,
                        det=line.det,
                        content=[content],
                    )],
                ))
        if mark_layout[1].lines:
            yield mark_layout

    def _extract_mark(self, line_contents: list[str | Reference]) -> tuple[Mark | str | None, str]:
        content = ""
        for item in line_contents:
            # 此阶段不会有 Reference 故可断言全为 str
            if isinstance(item, str):
                content += item
        content = content.lstrip()
        if not content:
            return None, ""

        matched = _START_PREFIX_PATTERN.match(content)
        if matched:
            prefix = matched.group(0)
            stars = prefix.strip()
            rest = content[matched.end():]
            return stars, rest
        else:
            mark = transform2mark(content[0])
            if mark is not None:
                return mark, content[1:]
            return None, content


import fitz

from typing import cast, Generator
from os import PathLike
from pathlib import Path
from PIL.Image import frombytes

from ..common import AssetHub
from ..metering import AbortedCheck
from .page_extractor import PageExtractorNode
from .types import Page, DeepSeekOCRModel


def pdf_pages_count(pdf_path: PathLike | str) -> int:
    with fitz.open(Path(pdf_path)) as document:
        return len(document)


class PageRefContext:
    def __init__(
            self,
            pdf_path: Path,
            extractor: PageExtractorNode,
            asset_hub: AssetHub,
            aborted: AbortedCheck,
        ) -> None:
        self._pdf_path = pdf_path
        self._extractor = extractor
        self._asset_hub = asset_hub
        self._aborted: AbortedCheck = aborted
        self._document: fitz.Document | None = None

    @property
    def pages_count(self) -> int:
        assert self._document is not None
        return len(self._document)

    def __enter__(self) -> "PageRefContext":
        assert self._document is None
        self._document = fitz.open(self._pdf_path)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if self._document is not None:
            self._document.close()
            self._document = None

    def __iter__(self) -> Generator["PageRef", None, None]:
        document = cast(fitz.Document, self._document)
        for i in range(len(document)):
            yield PageRef(
                document=document,
                page_index=i + 1,
                extractor=self._extractor,
                asset_hub=self._asset_hub,
                aborted=self._aborted,
            )

class PageRef:
    def __init__(
            self,
            document: fitz.Document,
            page_index: int,
            extractor: PageExtractorNode,
            asset_hub: AssetHub,
            aborted: AbortedCheck,
        ) -> None:
        self._document = document
        self._page_index = page_index
        self._extractor = extractor
        self._asset_hub = asset_hub
        self._aborted: AbortedCheck = aborted

    @property
    def page_index(self) -> int:
        return self._page_index

    def extract(
            self,
            model: DeepSeekOCRModel,
            includes_footnotes: bool,
            includes_raw_image: bool,
            plot_path: Path | None,
            max_tokens: int | None,
            max_output_tokens: int | None,
        ) -> Page:

        dpi = 300 # for scanned book pages
        default_dpi = 72
        matrix = fitz.Matrix(dpi / default_dpi, dpi / default_dpi)
        page = self._document.load_page(self._page_index - 1)
        pixmap = page.get_pixmap(matrix=matrix)
        image = frombytes("RGB", (pixmap.width, pixmap.height), pixmap.samples)

        return self._extractor.image2page(
            image=image,
            page_index=self._page_index,
            asset_hub=self._asset_hub,
            model_size=model,
            includes_footnotes=includes_footnotes,
            includes_raw_image=includes_raw_image,
            plot_path=plot_path,
            max_tokens=max_tokens,
            max_output_tokens=max_output_tokens,
            aborted=self._aborted,
        )


from dataclasses import dataclass
from typing import cast, Generator, Iterable
from xml.etree.ElementTree import Element

from ..common import indent, AssetRef, ASSET_TAGS
from .mark import Mark

@dataclass
class Chapter:
    title: "ParagraphLayout | None"
    layouts: list["ParagraphLayout | AssetLayout"]

@dataclass
class AssetLayout:
    page_index: int
    ref: AssetRef
    det: tuple[int, int, int, int]
    title: str | None
    content: str
    caption: str | None
    hash: str | None

@dataclass
class ParagraphLayout:
    ref: str
    lines: list["LineLayout"]

@dataclass
class LineLayout:
    page_index: int
    det: tuple[int, int, int, int]
    content: list["str | Reference"]

@dataclass
class Reference:
    page_index: int
    order: int
    mark: str | Mark
    layouts: list[AssetLayout | ParagraphLayout]

    @property
    def id(self) -> tuple[int, int]:
        return (self.page_index, self.order)

def search_references_in_chapter(chapter: Chapter) -> Generator[Reference, None, None]:
    seen: set[tuple[int, int]] = set()
    if chapter.title is not None:
        for line in chapter.title.lines:
            for part in line.content:
                if isinstance(part, Reference):
                    ref_id = part.id
                    if ref_id not in seen:
                        seen.add(ref_id)
                        yield part
    for layout in chapter.layouts:
        if isinstance(layout, ParagraphLayout):
            for line in layout.lines:
                for part in line.content:
                    if isinstance(part, Reference):
                        ref_id = part.id
                        if ref_id not in seen:
                            seen.add(ref_id)
                            yield part

def references_to_map(references: Iterable[Reference]) -> dict[tuple[int, int], int]:
    ref_id_to_number = {}
    for i, ref in enumerate(references, 1):
        ref_id_to_number[ref.id] = i
    return ref_id_to_number

def decode(element: Element) -> Chapter:
    references_el = element.find("references")
    references_map: dict[tuple[int, int], Reference] = {}
    if references_el is not None:
        for ref_el in references_el.findall("ref"):
            reference = _decode_reference(ref_el)
            references_map[reference.id] = reference

    title_el = element.find("title")
    title = _decode_paragraph(title_el, references_map) if title_el is not None else None
    body_el = element.find("body")
    if body_el is None:
        raise ValueError("<chapter> missing required <body> element")

    layouts: list[ParagraphLayout | AssetLayout] = []
    for child in list(body_el):
        tag = child.tag
        if tag == "asset":
            layouts.append(_decode_asset(child))
        elif tag == "paragraph":
            layouts.append(_decode_paragraph(child, references_map))

    return Chapter(
        title=title,
        layouts=layouts,
    )

def encode(chapter: Chapter) -> Element:
    root = Element("chapter")

    if chapter.title is not None:
        title_el = Element("title")
        title_el.set("ref", chapter.title.ref)
        _encode_line_elements(title_el, chapter.title.lines)
        root.append(title_el)

    body_el = Element("body")
    for layout in chapter.layouts:
        if isinstance(layout, AssetLayout):
            body_el.append(_encode_asset(layout))
        else:
            body_el.append(_encode_paragraph(layout))

    root.append(body_el)
    references = _collect_references(chapter)
    if references:
        references_el = Element("references")
        for ref in references:
            references_el.append(_encode_reference(ref))
        root.append(references_el)

    return indent(root, skip_tags=("line",))

def _decode_asset(element: Element) -> AssetLayout:
    ref_attr = element.get("ref")
    if ref_attr is None:
        raise ValueError("<asset> missing required attribute 'ref'")
    if ref_attr not in ASSET_TAGS:
        raise ValueError(f"<asset> attribute 'ref' must be one of {ASSET_TAGS}, got: {ref_attr}")
    page_index_attr = element.get("page_index")
    if page_index_attr is None:
        raise ValueError("<asset> missing required attribute 'page_index'")
    try:
        page_index = int(page_index_attr)
    except ValueError as e:
        raise ValueError(f"<asset> attribute 'page_index' must be int, got: {page_index_attr}") from e

    det_str = element.get("det")
    if det_str is None:
        raise ValueError("<asset> missing required attribute 'det'")
    det = _parse_det(det_str, context="<asset>@det")

    hash_value = element.get("hash")

    title_el = element.find("title")
    title = title_el.text if title_el is not None else None

    content_el = element.find("content")
    content = content_el.text if content_el is not None and content_el.text is not None else ""

    caption_el = element.find("caption")
    caption = caption_el.text if caption_el is not None else None

    return AssetLayout(
        page_index=page_index,
        ref=cast(AssetRef, ref_attr),
        det=det,
        title=title,
        content=content,
        caption=caption,
        hash=hash_value,
    )

def _encode_asset(layout: AssetLayout) -> Element:
    el = Element("asset")
    el.set("ref", layout.ref)
    el.set("page_index", str(layout.page_index))
    el.set("det", ",".join(map(str, layout.det)))
    if layout.hash is not None:
        el.set("hash", layout.hash)

    if layout.title is not None:
        title_el = Element("title")
        title_el.text = layout.title
        el.append(title_el)

    content_el = Element("content")
    content_el.text = layout.content
    el.append(content_el)

    if layout.caption is not None:
        caption_el = Element("caption")
        caption_el.text = layout.caption
        el.append(caption_el)

    return el

def _decode_paragraph(element: Element, references_map: dict[tuple[int, int], Reference] | None = None) -> ParagraphLayout:
    ref_attr = element.get("ref")
    if ref_attr is None:
        raise ValueError("<paragraph> missing required attribute 'ref'")

    lines = _decode_line_elements(element, context_tag="paragraph", references_map=references_map)

    return ParagraphLayout(ref=ref_attr, lines=lines)

def _encode_paragraph(layout: ParagraphLayout) -> Element:
    el = Element("paragraph")
    el.set("ref", layout.ref)
    _encode_line_elements(el, layout.lines)
    return el

def _parse_det(det_str: str, context: str) -> tuple[int, int, int, int]:
    try:
        det_list = list(map(int, det_str.split(",")))
    except Exception as e:
        raise ValueError(f"{context}: det must be comma-separated integers, got: {det_str}") from e
    if len(det_list) != 4:
        raise ValueError(f"{context}: det must have 4 values, got {len(det_list)}")
    return (det_list[0], det_list[1], det_list[2], det_list[3])

def _decode_line_elements(parent: Element, *, context_tag: str, references_map: dict[tuple[int, int], Reference] | None = None) -> list[LineLayout]:
    lines: list[LineLayout] = []
    for line_el in parent.findall("line"):
        page_index_attr = line_el.get("page_index")
        if page_index_attr is None:
            raise ValueError(f"<{context_tag}><line> missing required attribute 'page_index'")
        try:
            page_index = int(page_index_attr)
        except ValueError as e:
            raise ValueError(f"<{context_tag}><line> attribute 'page_index' must be int, got: {page_index_attr}") from e

        det_str = line_el.get("det")
        if det_str is None:
            raise ValueError(f"<{context_tag}><line> missing required attribute 'det'")
        det = _parse_det(det_str, context=f"<{context_tag}><line>@det")
        content: list[str | Reference] = []
        if line_el.text:
            content.append(line_el.text)

        for child in line_el:
            if child.tag == "ref":
                ref_id = child.get("id")
                if ref_id is None:
                    raise ValueError(f"<{context_tag}><line><ref> missing required attribute 'id'")

                try:
                    parts = ref_id.split('-')
                    if len(parts) != 2:
                        raise ValueError(f"<{context_tag}><line><ref> attribute 'id' must be in format 'page-order'")
                    ref_page_index = int(parts[0])
                    ref_order = int(parts[1])
                except ValueError as e:
                    raise ValueError(f"<{context_tag}><line><ref> attribute 'id' must contain valid integers") from e

                if references_map is not None:
                    ref_key = (ref_page_index, ref_order)
                    if ref_key in references_map:
                        content.append(references_map[ref_key])
                    else:
                        raise ValueError(f"<{context_tag}><line><ref> references undefined reference: {ref_id}")

            if child.tail:
                content.append(child.tail)

        lines.append(LineLayout(page_index=page_index, det=det, content=content))
    return lines

def _encode_line_elements(parent: Element, lines: list[LineLayout]) -> None:
    for line in lines:
        line_el = Element("line")
        line_el.set("page_index", str(line.page_index))
        line_el.set("det", ",".join(map(str, line.det)))
        has_elements = False
        for part in line.content:
            if isinstance(part, str):
                if not has_elements:
                    line_el.text = part
                else:
                    last_child = line_el[-1]
                    last_child.tail = part
            elif isinstance(part, Reference):
                ref_el = Element("ref")
                ref_el.set("id", f"{part.page_index}-{part.order}")
                line_el.append(ref_el)
                has_elements = True

        parent.append(line_el)

def _collect_references(chapter: Chapter) -> list[Reference]:
    references: list[Reference] = []
    seen: set[tuple[int, int]] = set()

    def collect_from_layouts(layouts: list[AssetLayout | ParagraphLayout]) -> None:
        for layout in layouts:
            if isinstance(layout, ParagraphLayout):
                for line in layout.lines:
                    for part in line.content:
                        if isinstance(part, Reference):
                            ref_id = part.id
                            if ref_id not in seen:
                                seen.add(ref_id)
                                references.append(part)

    if chapter.title is not None:
        for line in chapter.title.lines:
            for part in line.content:
                if isinstance(part, Reference):
                    ref_id = part.id
                    if ref_id not in seen:
                        seen.add(ref_id)
                        references.append(part)

    collect_from_layouts(chapter.layouts)
    references.sort(key=lambda ref: ref.id)

    return references

def _encode_reference(ref: Reference) -> Element:
    ref_el = Element("ref")
    ref_el.set("id", f"{ref.page_index}-{ref.order}")

    mark_el = Element("mark")
    mark_el.text = str(ref.mark)
    ref_el.append(mark_el)

    for layout in ref.layouts:
        if isinstance(layout, AssetLayout):
            ref_el.append(_encode_asset(layout))
        else:
            ref_el.append(_encode_paragraph(layout))

    return ref_el

def _decode_reference(element: Element) -> Reference:
    ref_id = element.get("id")
    if ref_id is None:
        raise ValueError("<references><ref> missing required attribute 'id'")

    try:
        parts = ref_id.split('-')
        if len(parts) != 2:
            raise ValueError("<references><ref> attribute 'id' must be in format 'page-order'")
        page_index = int(parts[0])
        order = int(parts[1])
    except ValueError as e:
        raise ValueError("<references><ref> attribute 'id' must contain valid integers") from e

    mark_el = element.find("mark")
    if mark_el is None or mark_el.text is None:
        raise ValueError("<references><ref> missing required <mark> element")
    mark_text = mark_el.text

    from .mark import transform2mark
    mark = transform2mark(mark_text)
    if mark is None:
        # 如果不是特殊标记，就使用原始字符串
        mark = mark_text

    layouts: list[AssetLayout | ParagraphLayout] = []
    for child in element:
        if child.tag == "mark":
            continue
        elif child.tag == "asset":
            layouts.append(_decode_asset(child))
        elif child.tag == "paragraph":
            layouts.append(_decode_paragraph(child, references_map=None))

    return Reference(
        page_index=page_index,
        order=order,
        mark=mark,
        layouts=layouts,
    )



from .generation import generate_chapter_files
from .chapter import decode, encode, search_references_in_chapter, references_to_map, Chapter, AssetLayout, AssetRef, ParagraphLayout, LineLayout
from .reference import Reference
from .mark import Mark, NumberClass, NumberStyle
from .language import is_chinese_char


from typing import Generator
from ..sequence import AssetLayout, ParagraphLayout
from ..sequence.chapter import Reference


def render_footnotes_section(references: list[Reference]) -> Generator[str, None, None]:
    if not references:
        return
    yield "\n\n---\n\n## References\n\n"
    for i, ref in enumerate(references, 1):
        yield f"[^{i}]: "
        yield from _render_reference_content(ref)
        yield "\n\n"

def _render_reference_content(ref: Reference):
    first = True
    for layout in ref.layouts:
        if isinstance(layout, AssetLayout):
            if layout.content:
                if not first:
                    yield " "
                yield layout.content.strip()
                first = False
        elif isinstance(layout, ParagraphLayout):
            for line in layout.lines:
                for part in line.content:
                    if isinstance(part, str):
                        if not first:
                            yield " "
                        yield part
                        first = False



from dataclasses import dataclass
from typing import Literal
from xml.etree.ElementTree import Element
from PIL.Image import Image

from ..common import indent


DeepSeekOCRModel = Literal["tiny", "small", "base", "large", "gundam"]

@dataclass
class Page:
    index: int
    image: Image | None
    body_layouts: list["PageLayout"]
    footnotes_layouts: list["PageLayout"]
    input_tokens: int
    output_tokens: int


@dataclass
class PageLayout:
    ref: str
    det: tuple[int, int, int, int]
    text: str
    hash: str | None


def decode(element: Element) -> Page:
    index = int(element.get("index", "0"))
    input_tokens = int(element.get("input_tokens", "0"))
    output_tokens = int(element.get("output_tokens", "0"))
    body_layouts = []
    body_element = element.find("body")
    if body_element is not None:
        for layout_element in body_element.findall("layout"):
            body_layouts.append(_decode_layout(layout_element))

    footnotes_layouts = []
    footnotes_element = element.find("footnotes")
    if footnotes_element is not None:
        for layout_element in footnotes_element.findall("layout"):
            footnotes_layouts.append(_decode_layout(layout_element))

    return Page(
        index=index,
        image=None,
        body_layouts=body_layouts,
        footnotes_layouts=footnotes_layouts,
        input_tokens=input_tokens,
        output_tokens=output_tokens
    )

def encode(page: Page) -> Element:
    page_element = Element("page")
    page_element.set("index", str(page.index))
    page_element.set("input_tokens", str(page.input_tokens))
    page_element.set("output_tokens", str(page.output_tokens))
    if page.body_layouts:
        body_element = Element("body")
        for layout in page.body_layouts:
            body_element.append(_encode_layout(layout))
        page_element.append(body_element)
    if page.footnotes_layouts:
        footnotes_element = Element("footnotes")
        for layout in page.footnotes_layouts:
            footnotes_element.append(_encode_layout(layout))
        page_element.append(footnotes_element)
    return indent(page_element)

def _decode_layout(element: Element) -> PageLayout:
    ref = element.get("ref", "")
    det_str = element.get("det", "0,0,0,0")
    det_list = list(map(int, det_str.split(",")))
    if len(det_list) != 4:
        raise ValueError(f"det must have 4 values, got {len(det_list)}")
    det = (det_list[0], det_list[1], det_list[2], det_list[3])
    text = element.text.strip() if element.text else ""
    hash_value = element.get("hash")
    return PageLayout(
        ref=ref,
        det=det,
        text=text,
        hash=hash_value
    )

def _encode_layout(layout: PageLayout) -> Element:
    layout_element = Element("layout")
    layout_element.set("ref", layout.ref)
    layout_element.set("det", ",".join(map(str, layout.det)))
    if layout.hash is not None:
        layout_element.set("hash", layout.hash)
    layout_element.text = layout.text
    return layout_element


import re

from typing import Generator, Iterable

from ..pdf import PageLayout
from ..common import ASSET_TAGS
from .chapter import ParagraphLayout, AssetLayout, LineLayout, Reference
from .language import is_latin_letter


TITLE_TAGS = ("title", "sub_title")

_ASSET_CAPTION_TAGS = tuple(f"{t}_caption" for t in ASSET_TAGS)

# to see https://github.com/opendatalab/MinerU/blob/fa1149cd4abf9db5e0f13e4e074cdb568be189f4/mineru/utils/span_pre_proc.py#L247
_LINE_STOP_FLAGS = (
    ".", "!", "?", "。", "！", "？", ")", "）", """, """, ":", "：", ";", "；",
    "]", "】", "}", "}", ">", "》", "、", ",", "，", "-", "—", "–",
)

_MARKDOWN_HEAD_PATTERN = re.compile(r"^#+\s+")
_LATEX_PATTERNS = [
    (re.compile(r"\\\["), re.compile(r"\\\]")),  # \[...\]
    (re.compile(r"\$\$"), re.compile(r"\$\$")),  # $$...$$
    (re.compile(r"\\\("), re.compile(r"\\\)")),  # \(...\)
    (re.compile(r"\$"), re.compile(r"\$")),      # $...$
]

_TABLE_PATTERN = re.compile(r"<table[^>]*>.*?</table>", re.IGNORECASE | re.DOTALL)


class Jointer:
    def __init__(self, layouts: Iterable[tuple[int, list[PageLayout]]]) -> None:
        self._layouts = layouts

    def execute(self) -> Generator[ParagraphLayout | AssetLayout, None, None]:
        last_page_para: ParagraphLayout | None = None
        for page_index, raw_layouts in self._layouts:
            layouts = self._transform_and_join_asset_layouts(page_index, raw_layouts)
            if not layouts:
                continue

            first_layout = layouts[0]
            if last_page_para and isinstance(first_layout, ParagraphLayout) and \
               self._can_merge_paragraphs(last_page_para, first_layout):
                last_page_para.lines.extend(first_layout.lines)
                del layouts[0]

            if not layouts:
                continue

            if last_page_para:
                _normalize_paragraph_content(last_page_para)
                yield last_page_para
                last_page_para = None

            for i in range(len(layouts) - 1):
                yield layouts[i]

            last_layout = layouts[-1]
            if last_layout:
                if isinstance(last_layout, ParagraphLayout):
                    last_page_para = last_layout
                else:
                    yield last_layout

        if last_page_para:
            _normalize_paragraph_content(last_page_para)
            yield last_page_para

    def _transform_and_join_asset_layouts(self, page_index, layouts: list[PageLayout]):
        last_asset: AssetLayout | None = None
        jointed_layouts: list[ParagraphLayout | AssetLayout] = []
        for layout in layouts:
            if layout.ref in ASSET_TAGS:
                if last_asset:
                    jointed_layouts.append(last_asset)
                last_asset = AssetLayout(
                    page_index=page_index,
                    ref=layout.ref,
                    det=layout.det,
                    title=None,
                    content=layout.text,
                    caption=None,
                    hash=layout.hash,
                )
            elif layout.ref in _ASSET_CAPTION_TAGS:
                if last_asset:
                    if last_asset.caption:
                        last_asset.caption += "\n" + layout.text
                    else:
                        last_asset.caption = layout.text
            else:
                if last_asset:
                    jointed_layouts.append(last_asset)
                    last_asset = None
                if layout.ref in TITLE_TAGS:
                    # 将 Markdown 标题前的 `##` 之类的符号删除，DeepSeek OCR 总会生成这种符号
                    layout.text = _MARKDOWN_HEAD_PATTERN.sub("", layout.text)

                jointed_layouts.append(ParagraphLayout(
                    ref=layout.ref,
                    lines=[LineLayout(
                        page_index=page_index,
                        det=layout.det,
                        content=[layout.text],
                    )],
                ))

        if last_asset:
            jointed_layouts.append(last_asset)

        for layout in jointed_layouts:
            if isinstance(layout, AssetLayout):
                if layout.ref == "equation":
                    _normalize_equation(layout)
                if layout.ref == "table":
                    _normalize_table(layout)
        return jointed_layouts

    # too see https://github.com/opendatalab/MinerU/blob/fa1149cd4abf9db5e0f13e4e074cdb568be189f4/mineru/backend/pipeline/para_split.py#L253
    def _can_merge_paragraphs(self, para1: ParagraphLayout, para2: ParagraphLayout) -> bool:
        if para1.ref != "text":
            return False
        if para1.ref != para2.ref:
            return False

        line1 = para1.lines[-1]
        line2 = para2.lines[0]
        det1, text1 = line1.det, line_text(line1)
        det2, text2 = line2.det, line_text(line2)

        if not text1 or not text2:
            return False

        text1_stripped = text1.rstrip()
        if not text1_stripped:
            return False

        text2_stripped = text2.lstrip()
        if not text2_stripped:
            return False

        # 条件1：前一个段落的末尾不以句尾符号结尾
        # 如果以句尾符号结尾，说明是完整段落，不应合并
        if text1_stripped.endswith(_LINE_STOP_FLAGS):
            return False

        layout_width1 = det1[2] - det1[0]
        layout_width2 = det2[2] - det2[0]

        # 条件2：两个段落的宽度相似
        # 差异不应超过较小宽度
        min_layout_width = min(layout_width1, layout_width2)
        if abs(layout_width1 - layout_width2) >= min_layout_width:
            return False

        first_char = text2_stripped[0]

        # 条件3：下一个段落的第一个字符不是数字
        # 如果以数字开头，可能是编号列表的新段落（如"1. xxx"）
        if first_char.isdigit():
            return False

        # 条件4：下一个段落的第一个字符不是大写字母
        # 如果以大写字母开头，可能是新段落的开始（特别是英文）
        if first_char.isupper():
            return False

        # 条件5：如果 para1 结尾是拉丁字母+"-"，para2 开头是拉丁字母，则允许合并（跨段单词拼接）
        if _is_splitted_word(text1, text2):
            return True

        return True

def _normalize_equation(layout: AssetLayout):
    content = layout.content
    if not content:
        return
    latex_start = -1
    latex_end = -1

    for start_pat, end_pat in _LATEX_PATTERNS:
        start_match = start_pat.search(content)
        if start_match:
            start_idx = start_match.start()
            end_match = end_pat.search(content[start_match.end():])
            if end_match:
                latex_start = start_idx
                latex_end = start_match.end() + end_match.end()
                break

    if latex_start >= 0 and latex_end > latex_start:
        _extract_and_split_content(layout, latex_start, latex_end)

def _normalize_table(layout: AssetLayout):
    content = layout.content
    if not content:
        return
    table_match = _TABLE_PATTERN.search(content)

    if table_match:
        table_start = table_match.start()
        table_end = table_match.end()
        _extract_and_split_content(layout, table_start, table_end)

def _normalize_paragraph_content(paragraph: ParagraphLayout):
    if len(paragraph.lines) < 2:
        return

    for i in range(1, len(paragraph.lines)):
        line1 = paragraph.lines[i - 1]
        line2 = paragraph.lines[i]
        content1 = line_text(line1).rstrip()
        content2 = line_text(line2).lstrip()

        if not _is_splitted_word(content1, content2):
            continue

        tail_end = 0
        for j in range(len(content2)):
            if is_latin_letter(content2[j]):
                tail_end = j + 1
            else:
                break

        line1.content[0] = content1[:-1] + content2[:tail_end]
        line2.content[0] = content2[tail_end:].lstrip()

    paragraph.lines = [
        line for line in paragraph.lines
        if line_text(line).strip()
    ]

def _extract_and_split_content(layout: AssetLayout, start: int, end: int):
    content = layout.content
    extracted = content[start:end]
    before = content[:start].strip()

    if before:
        if layout.title:
            layout.title = layout.title + "\n" + before
        else:
            layout.title = before

    after = content[end:].strip()
    if after:
        if layout.caption:
            layout.caption = layout.caption + "\n" + after
        else:
            layout.caption = after

    layout.content = extracted

def line_text(line: LineLayout) -> str:
    """
    从 LineLayout 中提取文本内容。
    支持处理混合的字符串和引用对象内容。
    """
    result_parts = []
    for part in line.content:
        if isinstance(part, str):
            result_parts.append(part)
        # 对于 Reference 对象，我们可以选择忽略或者转换为特定格式
        # 在 jointer 阶段，可能还需要原始文本，所以暂时转换为空字符串
        elif isinstance(part, Reference):
            # Reference 对象在这里暂时不处理，或者可以添加标记
            pass

    return "".join(result_parts)

def _is_splitted_word(text1: str, text2: str) -> bool:
    return (
        len(text1) >= 2 and text1[-1] == "-" and \
        is_latin_letter(text1[-2]) and \
        is_latin_letter(text2[0])
    )



from epub_generator import BookMeta, TableRender, LaTeXRender

from .pdf import pdf_pages_count, DeepSeekOCRModel, OCREvent, OCREventKind
from .transform import Transform, OCRTokensMetering
from .metering import AbortedCheck, InterruptedKind, InterruptedError
from .functions import transform_markdown, transform_epub, predownload_models


def is_latin_letter(char: str) -> bool:
    return char.isalpha() and ord(char) < 0x0370

def is_chinese_char(char: str) -> bool:
    if not char:
        return False
    code = ord(char)
    # 中文字符的 Unicode 范围
    return (
        0x4E00 <= code <= 0x9FFF or  # CJK统一汉字
        0x3400 <= code <= 0x4DBF or  # CJK统一汉字扩展A
        0x20000 <= code <= 0x2A6DF or  # CJK统一汉字扩展B
        0x2A700 <= code <= 0x2B73F or  # CJK统一汉字扩展C
        0x2B740 <= code <= 0x2B81F or  # CJK统一汉字扩展D
        0x2B820 <= code <= 0x2CEAF or  # CJK统一汉字扩展E
        0x3000 <= code <= 0x303F or  # CJK符号和标点
        0xFF00 <= code <= 0xFFEF
    )  # 全角ASCII、全角标点


from pathlib import Path
from typing import Literal
from epub_generator import (
    generate_epub,
    EpubData,
    BookMeta,
    TableRender,
    LaTeXRender,
    Chapter as ChapterRecord,
    ChapterGetter,
    TocItem,
    Text,
    Image,
    Formula,
    Footnote,
    Mark,
    TextKind,
)

from ..common import XMLReader
from ..metering import check_aborted, AbortedCheck
from ..sequence import (
    decode,
    search_references_in_chapter,
    references_to_map,
    Reference,
    Chapter,
    AssetLayout,
    ParagraphLayout,
)


def render_epub_file(
        chapters_path: Path,
        assets_path: Path,
        epub_path: Path,
        cover_path: Path | None,
        book_meta: BookMeta | None,
        lan: Literal["zh", "en"],
        table_render: TableRender,
        latex_render: LaTeXRender,
        aborted: AbortedCheck,
    ):

    chapters: XMLReader[Chapter] = XMLReader(
        prefix="chapter",
        dir_path=chapters_path,
        decode=decode,
    )
    references: list[Reference] = []
    for chapter in chapters.read():
        references.extend(search_references_in_chapter(chapter))

    references.sort(key=lambda ref: (ref.page_index, ref.order))
    ref_id_to_number = references_to_map(references)
    toc_items: list[TocItem] = []
    get_head: ChapterGetter | None = None

    for chapter in chapters.read():
        def get_chapter(ch=chapter):
            return _convert_chapter_to_epub(
                chapter=ch,
                ref_id_to_number=ref_id_to_number,
                assets_path=assets_path
            )
        if chapter.title is None:
            get_head = get_chapter
        else:
            title = _extract_chapter_title(chapter)
            toc_items.append(TocItem(
                title=title,
                get_chapter=get_chapter
            ))

    epub_data = EpubData(
        meta=book_meta,
        get_head=get_head,
        chapters=toc_items,
        cover_image_path=cover_path,
    )
    check_aborted(aborted)
    generate_epub(
        epub_data=epub_data,
        epub_file_path=epub_path,
        lan=lan,
        table_render=table_render,
        latex_render=latex_render,
        assert_not_aborted=lambda: check_aborted(aborted),
    )

def _extract_chapter_title(chapter: Chapter) -> str:
    if chapter.title is not None:
        text_parts = []
        for line in chapter.title.lines:
            for part in line.content:
                if isinstance(part, str):
                    text_parts.append(part)
        return " ".join(text_parts).strip() if text_parts else "Untitled"
    return "Untitled"

def _convert_chapter_to_epub(
    chapter: Chapter,
    ref_id_to_number: dict,
    assets_path: Path
) -> ChapterRecord:
    elements = []
    footnotes = []

    if chapter.title is not None:
        title_content = list(_render_paragraph_with_marks(
            layout=chapter.title,
            ref_id_to_number=ref_id_to_number),
        )
        if title_content:
            elements.append(Text(kind=TextKind.HEADLINE, content=title_content))

    for layout in chapter.layouts:
        if isinstance(layout, AssetLayout):
            asset_element = _convert_asset_to_epub(layout, assets_path)
            if asset_element:
                elements.append(asset_element)
        elif isinstance(layout, ParagraphLayout):
            paragraph_content = list(_render_paragraph_with_marks(
                layout=layout,
                ref_id_to_number=ref_id_to_number),
            )
            if paragraph_content:
                elements.append(Text(kind=TextKind.BODY, content=paragraph_content))

    chapter_refs = search_references_in_chapter(chapter)
    for ref in chapter_refs:
        footnotes.append(Footnote(
            id=ref_id_to_number.get(ref.id, 1),
            contents=list(_convert_reference_to_footnote_contents(
                ref=ref,
                assets_path=assets_path
            )),
        ))

    return ChapterRecord(elements=elements, footnotes=footnotes)

def _render_paragraph_with_marks(layout: ParagraphLayout, ref_id_to_number: dict):
    for line in layout.lines:
        for part in line.content:
            if isinstance(part, str):
                yield part
            elif isinstance(part, Reference):
                ref_number = ref_id_to_number.get(part.id, 1)
                yield Mark(id=ref_number)

def _convert_asset_to_epub(asset: AssetLayout, assets_path: Path):
    if asset.ref == "equation":
        latex_content = asset.content.strip()
        if not latex_content:
            return None

        if latex_content.startswith("\\[") and latex_content.endswith("\\]"):
            formula = latex_content[2:-2].strip()
        elif latex_content.startswith("\\(") and latex_content.endswith("\\)"):
            formula = latex_content[2:-2].strip()
        else:
            formula = latex_content

        return Formula(latex_expression=formula)

    elif asset.ref == "image":
        if asset.hash is None:
            return None

        image_file = assets_path / f"{asset.hash}.png"
        if not image_file.exists():
            return None

        alt_parts = []
        if asset.title:
            alt_parts.append(asset.title)
        if asset.caption:
            alt_parts.append(asset.caption)
        alt_text = " - ".join(alt_parts) if alt_parts else "image"

        return Image(path=image_file, alt_text=alt_text)

    elif asset.ref == "table":
        if asset.hash is None:
            return None

        table_file = assets_path / f"{asset.hash}.png"
        if not table_file.exists():
            return None

        alt_parts = []
        if asset.title:
            alt_parts.append(asset.title)
        if asset.caption:
            alt_parts.append(asset.caption)
        alt_text = " - ".join(alt_parts) if alt_parts else "table"

        return Image(path=table_file, alt_text=alt_text)

    return None

def _convert_reference_to_footnote_contents(ref: Reference, assets_path: Path):
    for layout in ref.layouts:
        if isinstance(layout, AssetLayout):
            asset_element = _convert_asset_to_epub(layout, assets_path)
            if asset_element:
                yield asset_element
        elif isinstance(layout, ParagraphLayout):
            content_parts: list[str | Mark] = []
            for line in layout.lines:
                for part in line.content:
                    if isinstance(part, str):
                        content_parts.append(part)

            if content_parts:
                yield Text(kind=TextKind.BODY, content=content_parts)


import hashlib
import uuid

from pathlib import Path
from typing import Literal
from PIL import Image


AssetRef = Literal["image", "table", "equation"]
ASSET_TAGS: tuple[AssetRef, ...] = ("image", "table", "equation")

class AssetHub:
    def __init__(self, asset_path: Path) -> None:
        self._asset_path = asset_path

    def clip(self, image: Image.Image, det: tuple[int, int, int, int]) -> str:
        cropped_image = image.crop(det)
        self._asset_path.mkdir(parents=True, exist_ok=True)
        temp_filename = f"{uuid.uuid4().hex}.png.temp"
        temp_path = self._asset_path / temp_filename
        try:
            cropped_image.save(temp_path, format="PNG")
            image_hash = self._calculate_file_hash(temp_path)
            target_path = self._asset_path / f"{image_hash}.png"
            if target_path.exists():
                temp_path.unlink()
                return image_hash
            temp_path.rename(target_path)
            return image_hash

        except Exception as e:
            if temp_path.exists():
                temp_path.unlink()
            raise e

    def _calculate_file_hash(self, file_path: Path) -> str:
        sha256 = hashlib.sha256()
        with open(file_path, "rb") as f:
            for chunk in iter(lambda: f.read(8192), b""):
                sha256.update(chunk)
        return sha256.hexdigest()


import re

from pathlib import Path
from typing import Generator, Callable, Generic, TypeVar
from xml.etree.ElementTree import Element, fromstring


T = TypeVar("T")


class XMLReader(Generic[T]):
    def __init__(self, prefix: str, dir_path: Path, decode: Callable[[Element], T]) -> None:
        dir_path = Path(dir_path)
        file_pattern = f"{prefix}_*.xml"
        regex = re.compile(rf"^{re.escape(prefix)}_(\d+)\.xml$")
        indexed_files: list[tuple[int, Path]] = []

        for p in dir_path.glob(file_pattern):
            m = regex.match(p.name)
            if not m:
                continue
            idx = int(m.group(1))
            indexed_files.append((idx, p))

        indexed_files.sort(key=lambda t: t[0])
        self._file_paths: list[Path] = [path for _, path in indexed_files]
        self._decode: Callable[[Element], T] = decode

    def read(self) -> Generator[T, None, None]:
        for xml_path in self._file_paths:
            try:
                root = fromstring(xml_path.read_text(encoding="utf-8"))
            except Exception as e:
                raise ValueError(f"Failed to parse XML file: {xml_path}") from e
            try:
                yield self._decode(root)
            except Exception as e:
                raise ValueError(f"Failed to decode from: {xml_path}") from e



from pathlib import Path
from shutil import copy2
from typing import Generator

from ..common import XMLReader
from ..metering import check_aborted, AbortedCheck
from ..sequence import (
    decode,
    is_chinese_char,
    search_references_in_chapter,
    references_to_map,
    Reference,
    Chapter,
    AssetLayout,
    ParagraphLayout,
)

from .footnotes import render_footnotes_section


def render_markdown_file(
        chapters_path: Path,
        assets_path: Path,
        output_path: Path,
        output_assets_path: Path,
        aborted: AbortedCheck,
    ):

    assets_ref_path = output_assets_path
    if not assets_ref_path.is_absolute():
        output_assets_path = output_path.parent / output_assets_path

    output_path.parent.mkdir(parents=True, exist_ok=True)
    output_assets_path.mkdir(parents=True, exist_ok=True)
    chapters: XMLReader[Chapter] = XMLReader(
        prefix="chapter",
        dir_path=chapters_path,
        decode=decode,
    )
    references: list[Reference] = []
    for chapter in chapters.read():
        references.extend(search_references_in_chapter(chapter))

    references.sort(key=lambda ref: (ref.page_index, ref.order))
    ref_id_to_number = references_to_map(references)

    with open(output_path, "w", encoding="utf-8") as f:
        for chapter in chapters.read():
            check_aborted(aborted)
            if chapter.title is not None:
                f.write("## ")
                for line in _render_paragraph_layout(chapter.title, ref_id_to_number):
                    f.write(line)
                f.write("\n\n")
            for layout in chapter.layouts:
                if isinstance(layout, AssetLayout):
                    asset_markdown = _render_asset(
                        asset=layout,
                        assets_path=assets_path,
                        output_assets_path=output_assets_path,
                        asset_ref_path=assets_ref_path,
                    )
                    if asset_markdown:
                        f.write(asset_markdown)
                elif isinstance(layout, ParagraphLayout):
                    for line in _render_paragraph_layout(layout, ref_id_to_number):
                        f.write(line)
                f.write("\n\n")

        check_aborted(aborted)
        for part in render_footnotes_section(references):
            f.write(part)

def _render_paragraph_layout(layout: ParagraphLayout, ref_id_to_number: dict) -> Generator[str, None, None]:
    last_char: str | None = None
    for line in _normalize_lines(layout, ref_id_to_number):
        if last_char is not None and (
            not is_chinese_char(last_char) or \
            not is_chinese_char(line[0])
        ):
            yield " "
        last_char = line[-1] if line else None
        yield line

def _normalize_lines(layout: ParagraphLayout, ref_id_to_number: dict) -> Generator[str, None, None]:
    for line_layout in layout.lines:
        line_content = _render_line_content(line_layout.content, ref_id_to_number)
        if line_content:
            for line in line_content.splitlines():
                if line:
                    yield line

def _render_line_content(content: list[str | Reference], ref_id_to_number: dict) -> str:
    result = []
    for part in content:
        if isinstance(part, str):
            result.append(part)
        elif isinstance(part, Reference):
            ref_number = ref_id_to_number.get(part.id, 1)
            result.append(f"[^{ref_number}]")
        else:
            result.append(str(part))
    return "".join(result)

def _render_asset(
    asset: AssetLayout,
    assets_path: Path,
    output_assets_path: Path,
    asset_ref_path: Path,
) -> str:

    alt_parts = []
    if asset.title:
        alt_parts.append(asset.title)
    if asset.caption:
        alt_parts.append(asset.caption)
    alt_text = " - ".join(alt_parts) if alt_parts else ""

    if asset.ref == "equation":
        latex_content = asset.content.strip()
        if not latex_content:
            return ""
        if latex_content.startswith("\\[") and latex_content.endswith("\\]"):
            # 块级公式 \[...\] -> $$...$$
            formula = latex_content[2:-2].strip()
            return f"$$\n{formula}\n$$\n"
        elif latex_content.startswith("\\(") and latex_content.endswith("\\)"):
            # 行内公式 \(...\) -> $...$
            formula = latex_content[2:-2].strip()
            return f"${formula}$\n"
        else:
            # 其他情况,假设为块级公式
            return f"$$\n{latex_content}\n$$\n"

    elif asset.ref in ("image", "table"):
        if asset.hash is None:
            return ""

        source_file = assets_path / f"{asset.hash}.png"
        if not source_file.exists():
            return ""

        target_file = output_assets_path / f"{asset.hash}.png"
        if not target_file.exists():
            copy2(source_file, target_file)

        if asset_ref_path.is_absolute():
            image_path = target_file
        else:
            image_path = asset_ref_path / f"{asset.hash}.png"

        # 使用 POSIX 风格路径(markdown 标准)
        image_path_str = str(image_path).replace("\\", "/")

        return f"![{alt_text}]({image_path_str})\n"

    return ""


from os import PathLike
from typing import Callable, Literal

from epub_generator import BookMeta, TableRender, LaTeXRender

from .pdf import OCR, OCREvent, DeepSeekOCRModel
from .transform import Transform
from .metering import AbortedCheck, OCRTokensMetering


def predownload_models(models_cache_path: PathLike | None = None) -> None:
    ocr = OCR(
        model_path=models_cache_path,
        local_only=False,
    )
    ocr.predownload()

def transform_markdown(
    pdf_path: PathLike,
    markdown_path: PathLike,
    markdown_assets_path: PathLike | None = None,
    analysing_path: PathLike | None = None,
    model: DeepSeekOCRModel = "gundam",
    models_cache_path: PathLike | None = None,
    local_only: bool = False,
    includes_footnotes: bool = False,
    generate_plot: bool = False,
    aborted: AbortedCheck = lambda: False,
    max_ocr_tokens: int | None = None,
    max_ocr_output_tokens: int | None = None,
    on_ocr_event: Callable[[OCREvent], None] = lambda _: None,
) -> OCRTokensMetering:

    return Transform(
        models_cache_path=models_cache_path,
        local_only=local_only,
    ).transform_markdown(
        pdf_path=pdf_path,
        markdown_path=markdown_path,
        markdown_assets_path=markdown_assets_path,
        analysing_path=analysing_path,
        model=model,
        includes_footnotes=includes_footnotes,
        generate_plot=generate_plot,
        aborted=aborted,
        max_ocr_tokens=max_ocr_tokens,
        max_ocr_output_tokens=max_ocr_output_tokens,
        on_ocr_event=on_ocr_event,
    )

def transform_epub(
    pdf_path: PathLike,
    epub_path: PathLike,
    analysing_path: PathLike | None = None,
    model: DeepSeekOCRModel = "gundam",
    models_cache_path: PathLike | None = None,
    local_only: bool = False,
    includes_cover: bool = True,
    includes_footnotes: bool = False,
    generate_plot: bool = False,
    book_meta: BookMeta | None = None,
    lan: Literal["zh", "en"] = "zh",
    table_render: TableRender = TableRender.HTML,
    latex_render: LaTeXRender = LaTeXRender.MATHML,
    aborted: AbortedCheck = lambda: False,
    max_ocr_tokens: int | None = None,
    max_ocr_output_tokens: int | None = None,
    on_ocr_event: Callable[[OCREvent], None] = lambda _: None,
) -> OCRTokensMetering:

    return Transform(
        models_cache_path=models_cache_path,
        local_only=local_only,
    ).transform_epub(
        pdf_path=pdf_path,
        epub_path=epub_path,
        analysing_path=analysing_path,
        model=model,
        includes_cover=includes_cover,
        includes_footnotes=includes_footnotes,
        generate_plot=generate_plot,
        book_meta=book_meta,
        lan=lan,
        table_render=table_render,
        latex_render=latex_render,
        aborted=aborted,
        max_ocr_tokens=max_ocr_tokens,
        max_ocr_output_tokens=max_ocr_output_tokens,
        on_ocr_event=on_ocr_event,
    )


from dataclasses import dataclass
import sys
import time

from typing import Container, Generator
from enum import auto, Enum
from pathlib import Path
from os import PathLike

from ..common import save_xml, AssetHub
from ..metering import check_aborted, AbortedCheck
from .page_extractor import PageExtractorNode
from .page_ref import PageRefContext
from .types import encode, DeepSeekOCRModel


class OCREventKind(Enum):
    START = auto()
    IGNORE = auto()
    SKIP = auto()
    COMPLETE = auto()

@dataclass
class OCREvent:
    kind: OCREventKind
    page_index: int
    total_pages: int
    cost_time_ms: int = 0
    input_tokens: int = 0
    output_tokens: int = 0

class OCR:
    def __init__(
            self,
            model_path: PathLike | None,
            local_only: bool,
        ) -> None:
        self._extractor = PageExtractorNode(
            model_path=model_path,
            local_only=local_only,
        )

    def predownload(self) -> None:
        self._extractor.download_models()

    def load_models(self) -> None:
        self._extractor.load_models()

    def recognize(
            self,
            pdf_path: Path,
            asset_path: Path,
            ocr_path: Path,
            model: DeepSeekOCRModel = "gundam",
            includes_footnotes: bool = False,
            plot_path: Path | None = None,
            cover_path: Path | None = None,
            aborted: AbortedCheck = lambda: False,
            page_indexes: Container[int] = range(1, sys.maxsize),
            max_tokens: int | None = None,
            max_output_tokens: int | None = None,
        ) -> Generator[OCREvent, None, None]:

        ocr_path.mkdir(parents=True, exist_ok=True)
        if plot_path is not None:
            plot_path.mkdir(parents=True, exist_ok=True)

        done_path = ocr_path / "done"
        did_ignore_any: bool = False
        if done_path.exists():
            return

        remain_tokens: int | None = max_tokens
        remain_output_tokens: int | None = max_output_tokens

        with PageRefContext(
            pdf_path=pdf_path,
            extractor=self._extractor,
            asset_hub=AssetHub(asset_path),
            aborted=aborted,
        ) as refs:
            pages_count = refs.pages_count
            for ref in refs:
                check_aborted(aborted)
                start_time = time.perf_counter()
                yield OCREvent(
                    kind=OCREventKind.START,
                    page_index=ref.page_index,
                    total_pages=pages_count,
                )
                if ref.page_index not in page_indexes:
                    elapsed_ms = int((time.perf_counter() - start_time) * 1000)
                    did_ignore_any = True
                    yield OCREvent(
                        kind=OCREventKind.IGNORE,
                        page_index=ref.page_index,
                        total_pages=pages_count,
                        cost_time_ms=elapsed_ms,
                    )
                    continue

                filename = f"page_{ref.page_index}.xml"
                file_path = ocr_path / filename

                if file_path.exists():
                    elapsed_ms = int((time.perf_counter() - start_time) * 1000)
                    yield OCREvent(
                        kind=OCREventKind.SKIP,
                        page_index=ref.page_index,
                        total_pages=pages_count,
                        cost_time_ms=elapsed_ms,
                    )
                else:
                    from doc_page_extractor import TokenLimitError
                    if remain_tokens is not None and remain_tokens <= 0:
                        raise TokenLimitError()
                    if remain_output_tokens is not None and remain_output_tokens <= 0:
                        raise TokenLimitError()

                    page = ref.extract(
                        model=model,
                        includes_footnotes=includes_footnotes,
                        includes_raw_image=(ref.page_index == 1),
                        plot_path=plot_path,
                        max_tokens=remain_tokens,
                        max_output_tokens=remain_output_tokens,
                    )
                    save_xml(encode(page), file_path)

                    if cover_path and page.image:
                        cover_path.parent.mkdir(parents=True, exist_ok=True)
                        page.image.save(cover_path, format="PNG")

                    elapsed_ms = int((time.perf_counter() - start_time) * 1000)

                    yield OCREvent(
                        kind=OCREventKind.COMPLETE,
                        page_index=ref.page_index,
                        total_pages=pages_count,
                        cost_time_ms=elapsed_ms,
                        input_tokens=page.input_tokens,
                        output_tokens=page.output_tokens,
                    )
                    if remain_tokens is not None:
                        remain_tokens -= page.input_tokens
                        remain_tokens -= page.output_tokens

                    if remain_output_tokens is not None:
                        remain_output_tokens -= page.output_tokens

        if not did_ignore_any:
            done_path.touch()


import re

from pathlib import Path
from os import PathLike
from PIL.Image import Image

from ..common import ASSET_TAGS, AssetHub
from ..metering import check_aborted, AbortedCheck
from .types import Page, PageLayout, DeepSeekOCRModel


class PageExtractorNode:
    def __init__(
        self,
        model_path: PathLike | None = None,
        local_only: bool = False,
    ) -> None:
        self._model_path: PathLike | None = model_path
        self._local_only: bool = local_only
        self._page_extractor = None

    def _get_page_extractor(self):
        if not self._page_extractor:
            # 尽可能推迟 doc-page-extractor 的加载时间
            from doc_page_extractor import PageExtractor
            self._page_extractor = PageExtractor(
                model_path=self._model_path,
                local_only=self._local_only,
            )
        return self._page_extractor

    def download_models(self) -> None:
        self._get_page_extractor().download_models()

    def load_models(self) -> None:
        self._get_page_extractor().load_models()

    def image2page(
            self,
            image: Image,
            page_index: int,
            asset_hub: AssetHub,
            model_size: DeepSeekOCRModel,
            includes_footnotes: bool,
            includes_raw_image: bool,
            plot_path: Path | None,
            max_tokens: int | None,
            max_output_tokens: int | None,
            aborted: AbortedCheck,
        ) -> Page:

        from doc_page_extractor import plot, ExtractionContext
        body_layouts: list[PageLayout] = []
        footnotes_layouts: list[PageLayout] = []
        raw_image: Image | None = None

        if includes_raw_image:
            raw_image = image
            image = image.copy()

        context = ExtractionContext(
            check_aborted=aborted,
            max_tokens=max_tokens,
            max_output_tokens=max_output_tokens,
        )
        for i, (image, layouts) in enumerate(self._get_page_extractor().extract(
            image=image,
            size=model_size,
            stages=2 if includes_footnotes else 1,
            context=context,
        )):
            for layout in layouts:
                ref = self._normalize_text(layout.ref)
                text = self._normalize_text(layout.text)
                hash: str | None = None
                if ref in ASSET_TAGS:
                    hash = asset_hub.clip(image, layout.det)
                page_layout = PageLayout(
                    ref=ref,
                    det=layout.det,
                    text=text,
                    hash=hash,
                )
                if i == 0:
                    body_layouts.append(page_layout)
                elif i == 1 and ref not in ASSET_TAGS:
                    footnotes_layouts.append(page_layout)

            check_aborted(aborted)
            if plot_path is not None:
                plot_file_path = plot_path / f"page_{page_index}_stage_{i + 1}.png"
                image = plot(image.copy(), layouts)
                image.save(plot_file_path, format="PNG")
                check_aborted(aborted)

        return Page(
            index=page_index,
            image=raw_image,
            body_layouts=body_layouts,
            footnotes_layouts=footnotes_layouts,
            input_tokens=context.input_tokens,
            output_tokens=context.output_tokens,
        )

    def _normalize_text(self, text: str | None) -> str:
        if text is None:
            return ""
        text = re.sub(r"\s+", " ", text)
        return text.strip()



import re
import io

from dataclasses import dataclass
from enum import Enum, auto
from typing import Generator, Iterable


class NumberClass(Enum):
    ROMAN_NUMERAL = auto()                   # 罗马数字
    LOWERCASE_ROMAN_NUMERAL = auto()         # 小写罗马数字
    CIRCLED_NUMBER = auto()                  # 带圆圈的数字
    PARENTHESIZED_CHINESE = auto()           # 括号中的汉字
    CIRCLED_CHINESE = auto()                 # 带圆圈的汉字
    BLACK_CIRCLED_NUMBER = auto()            # 黑色圆圈数字
    UNBOUNDED_NUMBER = auto()                # 无包围数字

class NumberStyle(Enum):
    ROMAN_NUMERAL = auto()                   # 罗马数字
    LOWERCASE_ROMAN_NUMERAL = auto()         # 小写罗马数字
    CIRCLED_NUMBER = auto()                  # 带圆圈的数字
    DOUBLE_CIRCLED_NUMBER = auto()           # 双圈数字
    CIRCLED_SANS_SERIF_NUMBER = auto()       # 带圆圈的无衬线数字
    BLACK_CIRCLED_SANS_SERIF_NUMBER = auto() # 黑色圆圈无衬线数字
    BLACK_CIRCLED_NUMBER = auto()            # 黑色圆圈数字
    PARENTHESIZED_CHINESE = auto()           # 括号中的汉字
    CIRCLED_CHINESE = auto()                 # 带圆圈的汉字
    FULL_WIDTH_NUMBER = auto()               # 全角数字
    MATHEMATICAL_BOLD_NUMBER = auto()        # 数学粗体数字
    ARTISTIC_BOLD_NUMBER = auto()            # 艺术粗体数字
    OUTLINED_BOLD_NUMBER = auto()            # 描边的粗体数字
    SUBSCRIPT_NUMBER = auto()                # 带角标的数字

@dataclass
class Mark:
    number: int
    char: str
    clazz: NumberClass
    style: NumberStyle

    def __str__(self) -> str:
        return self.char

    def __hash__(self):
        return hash((self.clazz, self.number))

    def __eq__(self, other) -> bool:
        if not isinstance(other, Mark):
            return False
        if self.clazz != other.clazz:
            return False
        if self.number != other.number:
            return False
        return True

def samples(number_style: NumberStyle, count: int) -> str:
    if count <= 1:
        raise ValueError("Count must be greater than 1")
    half_count = count // 2
    number_styles = _number_marks.styles.get(number_style, None)
    if number_styles is None:
        raise ValueError(f"Invalid number style: {number_style.name}")
    buffer = io.StringIO()
    for char in number_styles[:half_count]:
        buffer.write(char)
    buffer.write("...")
    for char in number_styles[-half_count:]:
        buffer.write(char)
    return buffer.getvalue()

def transform2mark(raw_char: str) -> Mark | None:
    gotten = _number_marks.marks.get(raw_char, None)
    if gotten is None:
        return None
    return Mark(
            number=gotten.number,
            char=gotten.char,
            clazz=gotten.clazz,
            style=gotten.style
    )

def search_marks(text: str) -> Generator[Mark | str, None, None]:
    for part in re.split(_number_marks.pattern, text):
        mark = transform2mark(part)
        if mark is None:
            yield part
        else:
            yield mark

class _NumberMarks:
    def __init__(self, styles: Iterable[tuple[NumberClass, NumberStyle, Iterable[tuple[int, str]]]]):
        self.marks: dict[str, Mark] = {}
        self.styles: dict[NumberStyle, list[str]] = {}
        for clazz, style, marks in styles:
            for number, mark in marks:
                self.marks[mark] = Mark(number, mark, clazz, style)
                self.styles[style] = [char for _, char in sorted(marks, key=lambda x: x[0])]

        self.pattern: re.Pattern = re.compile(
            r"([" + "".join(sorted(list(self.marks.keys()))) + r"])"
        )

# some of they are from https://tw.piliapp.com/symbol/number/
_number_marks = _NumberMarks((
    (NumberClass.ROMAN_NUMERAL, NumberStyle.ROMAN_NUMERAL, (
        (1, "Ⅰ"), (2, "Ⅱ"), (3, "Ⅲ"), (4, "Ⅳ"), (5, "Ⅴ"),
        (6, "Ⅵ"), (7, "Ⅶ"), (8, "Ⅷ"), (9, "Ⅸ"), (10, "Ⅹ"),
        (11, "Ⅺ"), (12, "Ⅻ"),
    )),
    (NumberClass.LOWERCASE_ROMAN_NUMERAL, NumberStyle.LOWERCASE_ROMAN_NUMERAL, (
        (1, "ⅰ"), (2, "ⅱ"), (3, "ⅲ"), (4, "ⅳ"), (5, "ⅴ"),
        (6, "ⅵ"), (7, "ⅶ"), (8, "ⅷ"), (9, "ⅸ"), (10, "ⅹ"),
        (11, "ⅺ"), (12, "ⅻ"),
    )),
    (NumberClass.CIRCLED_NUMBER, NumberStyle.CIRCLED_NUMBER, (
        (0, "⓪"), (1, "①"), (2, "②"), (3, "③"), (4, "④"),
        (5, "⑤"), (6, "⑥"), (7, "⑦"), (8, "⑧"), (9, "⑨"),
        (10, "⑩"), (11, "⑪"), (12, "⑫"), (13, "⑬"), (14, "⑭"),
        (15, "⑮"), (16, "⑯"), (17, "⑰"), (18, "⑱"), (19, "⑲"),
        (20, "⑳"), (21, "㉑"), (22, "㉒"), (23, "㉓"), (24, "㉔"),
        (25, "㉕"), (26, "㉖"), (27, "㉗"), (28, "㉘"), (29, "㉙"),
        (30, "㉚"), (31, "㉛"), (32, "㉜"), (33, "㉝"), (34, "㉞"),
        (35, "㉟"), (36, "㊱"), (37, "㊲"), (38, "㊳"), (39, "㊴"),
        (40, "㊵"), (41, "㊶"), (42, "㊷"), (43, "㊸"), (44, "㊹"),
        (45, "㊺"), (46, "㊻"), (47, "㊼"), (48, "㊽"), (49, "㊾"),
        (50, "㊿")
    )),
    (NumberClass.CIRCLED_NUMBER, NumberStyle.DOUBLE_CIRCLED_NUMBER, (
        (0, "⓵"), (1, "⓶"), (2, "⓷"), (3, "⓸"), (4, "⓹"),
        (5, "⓺"), (6, "⓻"), (7, "⓼"), (8, "⓽"), (9, "⓾"),
    )),
    (NumberClass.CIRCLED_NUMBER, NumberStyle.CIRCLED_SANS_SERIF_NUMBER, (
        (1, "➀"), (2, "➁"), (3, "➂"), (4, "➃"), (5, "➄"),
        (6, "➅"), (7, "➆"), (8, "➇"), (9, "➈"), (10, "➉"),
    )),
    (NumberClass.BLACK_CIRCLED_NUMBER, NumberStyle.BLACK_CIRCLED_SANS_SERIF_NUMBER, (
        (1, "➊"), (2, "➋"), (3, "➌"), (4, "➍"), (5, "➎"),
        (6, "➏"), (7, "➐"), (8, "➑"), (9, "➒"), (10, "➓"),
    )),
    (NumberClass.BLACK_CIRCLED_NUMBER, NumberStyle.BLACK_CIRCLED_NUMBER, (
        (0, "⓿"), (1, "❶"), (2, "❷"), (3, "❸"), (4, "❹"),
        (5, "❺"), (6, "❻"), (7, "❼"), (8, "❽"), (9, "❾"),
        (10, "❿"), (11, "⓫"), (12, "⓬"), (13, "⓭"), (14, "⓮"),
        (15, "⓯"), (16, "⓰"), (17, "⓱"), (18, "⓲"), (19, "⓳"),
        (20, "⓴"),
    )),
    (NumberClass.PARENTHESIZED_CHINESE, NumberStyle.PARENTHESIZED_CHINESE, (
        (1, "㈠"), (2, "㈡"), (3, "㈢"), (4, "㈣"), (5, "㈤"),
        (6, "㈥"), (7, "㈦"), (8, "㈧"), (9, "㈨"), (10, "㈩"),
    )),
    (NumberClass.CIRCLED_CHINESE, NumberStyle.CIRCLED_CHINESE, (
        (1, "㊀"), (2, "㊁"), (3, "㊂"), (4, "㊃"), (5, "㊄"),
        (6, "㊅"), (7, "㊆"), (8, "㊇"), (9, "㊈"), (10, "㊉"),
    )),
    (NumberClass.UNBOUNDED_NUMBER, NumberStyle.FULL_WIDTH_NUMBER, (
        (0, "０"), (1, "１"), (2, "２"), (3, "３"), (4, "４"),
        (5, "５"), (6, "６"), (7, "７"), (8, "８"), (9, "９"),
    )),
    (NumberClass.UNBOUNDED_NUMBER, NumberStyle.MATHEMATICAL_BOLD_NUMBER, (
        (0, "𝟬"), (1, "𝟭"), (2, "𝟮"), (3, "𝟯"), (4, "𝟰"),
        (5, "𝟱"), (6, "𝟲"), (7, "𝟳"), (8, "𝟴"), (9, "𝟵"),
    )),
    (NumberClass.UNBOUNDED_NUMBER, NumberStyle.ARTISTIC_BOLD_NUMBER, (
        (0, "𝟎"), (1, "𝟏"), (2, "𝟐"), (3, "𝟑"), (4, "𝟒"),
        (5, "𝟓"), (6, "𝟔"), (7, "𝟕"), (8, "𝟖"), (9, "𝟗"),
    )),
    (NumberClass.UNBOUNDED_NUMBER, NumberStyle.OUTLINED_BOLD_NUMBER, (
        (0, "𝟘"), (1, "𝟙"), (2, "𝟚"), (3, "𝟛"), (4, "𝟜"),
        (5, "𝟝"), (6, "𝟞"), (7, "𝟟"), (8, "𝟠"), (9, "𝟡"),
    )),
    (NumberClass.UNBOUNDED_NUMBER, NumberStyle.SUBSCRIPT_NUMBER, (
        (0, "🄁"), (1, "🄂"), (2, "🄃"), (3, "🄄"), (4, "🄅"),
        (5, "🄆"), (6, "🄇"), (7, "🄈"), (8, "🄉"), (9, "🄊"),
    )),
))



from os import PathLike
from pathlib import Path
import tempfile


class EnsureFolder:
    def __init__(self, path: PathLike | str | None):
        self._path = Path(path) if path is not None else None
        self._temp: tempfile.TemporaryDirectory | None = None

    def __enter__(self) -> Path:
        if self._path is None:
            self._temp = tempfile.TemporaryDirectory()
            self._path = Path(self._temp.name)
        else:
            self._path.mkdir(parents=True, exist_ok=True)
        return self._path

    def __exit__(self, exc_type, exc, tb) -> bool:
        if self._temp is not None:
            try:
                self._temp.cleanup()
            finally:
                self._temp = None
        return False


from pathlib import Path
from typing import Generator

from ..common import save_xml, XMLReader
from ..pdf import decode, Page
from .jointer import TITLE_TAGS, Jointer
from .chapter import encode, Reference, Chapter, AssetLayout, ParagraphLayout, LineLayout
from .reference import References
from .mark import search_marks, Mark


def generate_chapter_files(pages_path: Path, chapters_path: Path):
    chapters_path.mkdir(parents=True, exist_ok=True)
    for chapter_file in chapters_path.glob("chapter_*.xml"):
        chapter_file.unlink()

    for i, chapter in enumerate(_generate_chapters(pages_path)):
        chapter_file = chapters_path / f"chapter_{i + 1}.xml"
        chapter_element = encode(chapter)
        save_xml(chapter_element, chapter_file)

def _generate_chapters(pages_path: Path):
    chapter: Chapter | None = None
    for layout in _extract_body_layouts(pages_path):
        if isinstance(layout, ParagraphLayout) and layout.ref in TITLE_TAGS:
            if chapter:
                yield chapter
            chapter = Chapter(title=layout, layouts=[])
        else:
            if chapter is None:
                chapter = Chapter(title=None, layouts=[])
            chapter.layouts.append(layout)

    if chapter:
        yield chapter

def _extract_body_layouts(pages_path: Path):
    pages: XMLReader[Page] = XMLReader(
        prefix="page",
        dir_path=pages_path,
        decode=decode,
    )
    body_jointer = Jointer(((p.index, p.body_layouts) for p in pages.read()))
    footnotes_jointer = Jointer(((p.index, p.footnotes_layouts) for p in pages.read()))
    references_generator = _extract_page_references(footnotes_jointer)
    current_references: References | None = next(references_generator, None)

    def get_references(page_index: int) -> References | None:
        nonlocal current_references
        while current_references is not None and current_references.page_index < page_index:
            current_references = next(references_generator, None)
        if current_references is not None and current_references.page_index == page_index:
            return current_references
        return None

    for layout in body_jointer.execute():
        if isinstance(layout, ParagraphLayout):
            for line in layout.lines:
                references = get_references(line.page_index)
                if references:
                    line.content = list(_line_parts_after_replace_references(references, line))
        yield layout

def _extract_page_references(jointer: Jointer) -> Generator[References, None, None]:
    last_page_index: int = -1
    layout_buffer: list[AssetLayout | ParagraphLayout] = []

    for layout in jointer.execute():
        page_index = _page_index_from_layout(layout)
        if page_index != last_page_index:
            if layout_buffer:
                yield References(
                    page_index=last_page_index,
                    layouts=layout_buffer,
                )
            last_page_index = page_index
            layout_buffer = []
        layout_buffer.append(layout)

    if layout_buffer:
        yield References(
            page_index=last_page_index,
            layouts=layout_buffer,
        )

def _page_index_from_layout(layout: AssetLayout | ParagraphLayout) -> int:
    if isinstance(layout, ParagraphLayout):
        return layout.lines[0].page_index
    elif isinstance(layout, AssetLayout):
        return layout.page_index
    else:
        raise ValueError("Unknown layout type")


def _line_parts_after_replace_references(references: References, line: LineLayout) -> Generator[str | Reference, None, None]:
    text_buffer: list[str] = []
    for part in _search_mark_and_split_line(references, line.content):
        if isinstance(part, str):
            text_buffer.append(part)
        elif text_buffer:
            yield "".join(text_buffer)
            text_buffer = []
            yield part
    if text_buffer:
        yield "".join(text_buffer)

def _search_mark_and_split_line(references: References, parts: list[str | Reference]):
    for part in parts:
        if isinstance(part, Reference):
            yield part
        else:
            for item in search_marks(part):
                reference: Reference | None = None
                if isinstance(item, Mark):
                    reference = references.get(item)
                if reference:
                    yield reference
                else:
                    yield str(item)


from os import PathLike
from pathlib import Path
from typing import Callable, Literal

from epub_generator import BookMeta, TableRender, LaTeXRender

from .common import EnsureFolder
from .pdf import OCR, OCREvent, DeepSeekOCRModel
from .sequence import generate_chapter_files
from .markdown import render_markdown_file
from .epub import render_epub_file
from .metering import to_interrupted_error, AbortedCheck, OCRTokensMetering


class Transform:
    def __init__(
            self,
            models_cache_path: PathLike | None = None,
            local_only: bool = False,
        ) -> None:
        self._ocr: OCR = OCR(
            model_path=models_cache_path,
            local_only=local_only,
        )

    def predownload(self) -> None:
        self._ocr.predownload()

    def load_models(self) -> None:
        self._ocr.load_models()

    def transform_markdown(
        self,
        pdf_path: PathLike | str,
        markdown_path: PathLike | str,
        markdown_assets_path: PathLike | str | None = None,
        analysing_path: PathLike | str | None = None,
        model: DeepSeekOCRModel = "gundam",
        includes_footnotes: bool = False,
        generate_plot: bool = False,
        aborted: AbortedCheck = lambda: False,
        max_ocr_tokens: int | None = None,
        max_ocr_output_tokens: int | None = None,
        on_ocr_event: Callable[[OCREvent], None] = lambda _: None,
    ) -> OCRTokensMetering: # pyright: ignore[reportReturnType]

        if markdown_assets_path is None:
            markdown_assets_path = Path(".") / "assets"
        else:
            markdown_assets_path = Path(markdown_assets_path)
        try:
            with EnsureFolder(analysing_path) as analysing_path:
                asserts_path, chapters_path, _, metering = self._extract_from_pdf(
                    pdf_path=Path(pdf_path),
                    analysing_path=analysing_path,
                    model=model,
                    includes_cover=False,
                    includes_footnotes=includes_footnotes,
                    generate_plot=generate_plot,
                    aborted=aborted,
                    max_tokens=max_ocr_tokens,
                    max_output_tokens=max_ocr_output_tokens,
                    on_ocr_event=on_ocr_event,
                )
                render_markdown_file(
                    chapters_path=chapters_path,
                    assets_path=asserts_path,
                    output_path=Path(markdown_path),
                    output_assets_path=markdown_assets_path,
                    aborted=aborted,
                )
                return metering

        except Exception as raw_error:
            error = to_interrupted_error(raw_error)
            if error:
                raise error from raw_error
            else:
                raise RuntimeError(f"transform {pdf_path} to markdown failed") from raw_error

    def transform_epub(
        self,
        pdf_path: PathLike | str,
        epub_path: PathLike | str,
        analysing_path: PathLike | str | None = None,
        model: DeepSeekOCRModel = "gundam",
        includes_cover: bool = True,
        includes_footnotes: bool = False,
        generate_plot: bool = False,
        book_meta: BookMeta | None = None,
        lan: Literal["zh", "en"] = "zh",
        table_render: TableRender = TableRender.HTML,
        latex_render: LaTeXRender = LaTeXRender.MATHML,
        aborted: AbortedCheck = lambda: False,
        max_ocr_tokens: int | None = None,
        max_ocr_output_tokens: int | None = None,
        on_ocr_event: Callable[[OCREvent], None] = lambda _: None,
    ) -> OCRTokensMetering:  # pyright: ignore[reportReturnType]

        try:
            with EnsureFolder(analysing_path) as analysing_path:
                asserts_path, chapters_path, cover_path, metering = self._extract_from_pdf(
                    pdf_path=Path(pdf_path),
                    analysing_path=analysing_path,
                    model=model,
                    includes_cover=includes_cover,
                    includes_footnotes=includes_footnotes,
                    generate_plot=generate_plot,
                    aborted=aborted,
                    max_tokens=max_ocr_tokens,
                    max_output_tokens=max_ocr_output_tokens,
                    on_ocr_event=on_ocr_event,
                )
                render_epub_file(
                    chapters_path=chapters_path,
                    assets_path=asserts_path,
                    epub_path=Path(epub_path),
                    book_meta=book_meta,
                    lan=lan,
                    cover_path=cover_path,
                    table_render=table_render,
                    latex_render=latex_render,
                    aborted=aborted,
                )
                return metering

        except Exception as raw_error:
            error = to_interrupted_error(raw_error)
            if error:
                raise error from raw_error
            else:
                raise RuntimeError(f"transform {pdf_path} to epub failed") from raw_error

    def _extract_from_pdf(
        self,
        pdf_path: Path,
        analysing_path: Path,
        model: DeepSeekOCRModel,
        includes_cover: bool,
        includes_footnotes: bool,
        generate_plot: bool,
        aborted: AbortedCheck,
        max_tokens: int | None,
        max_output_tokens: int | None,
        on_ocr_event: Callable[[OCREvent], None],
    ):

        asserts_path = analysing_path / "assets"
        pages_path = analysing_path / "ocr"
        chapters_path = analysing_path / "chapters"

        cover_path: Path | None = None
        plot_path: Path | None = None
        if includes_cover:
            cover_path = analysing_path / "cover.png"
        if generate_plot:
            plot_path = analysing_path / "plots"

        metering = OCRTokensMetering(
            input_tokens=0,
            output_tokens=0,
        )
        for event in self._ocr.recognize(
            pdf_path=pdf_path,
            asset_path=asserts_path,
            ocr_path=pages_path,
            model=model,
            includes_footnotes=includes_footnotes,
            plot_path=plot_path,
            cover_path=cover_path,
            aborted=aborted,
            max_tokens=max_tokens,
            max_output_tokens=max_output_tokens,
        ):
            on_ocr_event(event)
            metering.input_tokens += event.input_tokens
            metering.output_tokens += event.output_tokens

        generate_chapter_files(
            pages_path=pages_path,
            chapters_path=chapters_path,
        )
        return asserts_path, chapters_path, cover_path, metering


from pathlib import Path
from typing import Container
from xml.etree.ElementTree import tostring, Element


def indent(elem: Element, level: int = 0, skip_tags: Container[str] = ()) -> Element:
    indent_str = "  " * level
    next_indent_str = "  " * (level + 1)

    if elem.tag in skip_tags:
        if level > 0 and (not elem.tail or not elem.tail.strip()):
            elem.tail = "\n" + indent_str
        return elem

    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = "\n" + next_indent_str
        for i, child in enumerate(elem):
            indent(child, level + 1, skip_tags)
            if i < len(elem) - 1:
                child.tail = "\n" + next_indent_str
            else:
                child.tail = "\n" + indent_str
    elif level > 0 and (not elem.tail or not elem.tail.strip()):
        elem.tail = "\n" + indent_str
    return elem

def save_xml(element: Element, file_path: Path) -> None:
    # 使用临时文件确保写入的原子性
    xml_string = tostring(element, encoding="unicode")
    temp_path = file_path.with_suffix(".xml.tmp")
    try:
        with open(temp_path, "w", encoding="utf-8") as f:
            f.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            f.write(xml_string)
        temp_path.replace(file_path)
    except Exception as err:
        if temp_path.exists():
            temp_path.unlink()
        raise err


from dataclasses import dataclass
from enum import auto, Enum
from typing import Callable


AbortedCheck = Callable[[], bool]

def check_aborted(aborted_check: AbortedCheck) -> None:
    if aborted_check():
        from doc_page_extractor import AbortError
        raise AbortError()

@dataclass
class OCRTokensMetering:
    input_tokens: int
    output_tokens: int


class InterruptedKind(Enum):
    ABORT = auto()
    TOKEN_LIMIT_EXCEEDED = auto()


# 不可直接用 doc-page-extractor 的 Error，该库的一切都是懒加载，若暴露，则无法懒加载
class InterruptedError(Exception):
    """Raised when the operation is interrupted by the user."""
    def __init__(self, metering: OCRTokensMetering) -> None:
        super().__init__()
        self._kind: InterruptedKind
        self._metering: OCRTokensMetering = metering

def to_interrupted_error(error: Exception) -> InterruptedError | None:
    from doc_page_extractor import AbortError, TokenLimitError, ExtractionAbortedError
    if isinstance(error, ExtractionAbortedError):
        kind: InterruptedKind | None = None
        if isinstance(error, AbortError):
            kind = InterruptedKind.ABORT
        elif isinstance(error, TokenLimitError):
            kind = InterruptedKind.TOKEN_LIMIT_EXCEEDED
        if kind is not None:
            return InterruptedError(OCRTokensMetering(
                input_tokens=error.input_tokens,
                output_tokens=error.output_tokens,
            ))
    return None